import { ɵɵdefineInjectable, Injectable } from '@angular/core';
import { of, interval, BehaviorSubject, forkJoin, throwError, concat } from 'rxjs';
import { map, scan, takeWhile, tap, catchError, concatMap, switchMap, last, repeat } from 'rxjs/operators';

class Battleship_performanceError extends Error {
    constructor(message) {
        super(message);
        this.name = "Battleship_performanceError";
    }
}

class ShotClass {
    constructor(oppShips) {
        this._shots = [];
        this._oppShips = oppShips;
        this._shotOpti = new Map();
    }
    get shots() {
        return this._shots;
    }
    shotBefore(x) {
        return this._shotOpti.has(x);
    }
    shoot(x, s) {
        if (this.shotBefore(x)) {
            return false;
        }
        for (let j of this._oppShips.shipman.ships) {
            for (let i of j.pos) {
                if (x === i) {
                    let tmp = Object.assign({}, s.value);
                    tmp[j.length] = tmp[j.length] - 1;
                    s.next(tmp); //to update the score
                    this._shots.push({ hit: true, pos: i, shipLen: j.length });
                    this._shotOpti.set(i, i);
                    return true;
                }
            }
        }
        this._shots.push({ hit: false, pos: x });
        this._shotOpti.set(x, x);
        return true;
    }
}

const randomNum = (x) => Math.floor(Math.random() * x) + 1;

class PlayerShotClass extends ShotClass {
    constructor(oppShips) {
        super(oppShips);
    }
    shoot(x, scoreBoard) {
        x = x = randomNum(this._oppShips.gridSize * this._oppShips.gridSize);
        while (!super.shoot(x, scoreBoard))
            x = randomNum(this._oppShips.gridSize * this._oppShips.gridSize);
        return true;
    }
}

class ShipClass {
    constructor(shipman, gridSize, shipsNo) {
        //need to set shipsNo and gridSize here  also need to set getters and setters
        this._shipman = shipman;
        this._gridSize = gridSize;
        this._shipsNo = shipsNo;
    }
    shipBefore(x) {
        return this._shipman.allPos.has(x);
    }
    get gridSize() {
        return this._gridSize;
    }
    get shipsNo() {
        return this._shipsNo;
    }
    add(curr) {
        if (this._shipman.ships.length === 0) {
            let pos = [];
            pos.push(curr);
            this._shipman.allPos.set(curr, curr);
            this._shipman.ships.push({ pos: pos, length: this._shipsNo, posibleDirection: "all" /* all */ });
        }
        else {
            if (this._shipman.ships[this._shipman.ships.length - 1].pos.length < this._shipsNo - this._shipman.ships.length + 1) {
                let len = this._shipsNo - this._shipman.ships.length + 1;
                let d = this.checkValidity(curr, len, this._shipman.ships[this._shipman.ships.length - 1]);
                if (d !== "no" /* no */) {
                    this._shipman.allPos.set(curr, curr);
                    this._shipman.ships[this._shipman.ships.length - 1].length = len;
                    this._shipman.ships[this._shipman.ships.length - 1].pos.push(curr);
                    this._shipman.ships[this._shipman.ships.length - 1].posibleDirection = d;
                }
            }
            else {
                //first time working on a new ship
                let len = this._shipsNo - this._shipman.ships.length;
                let d = this.checkValidity(curr, len);
                if (d !== "no" /* no */) {
                    this._shipman.allPos.set(curr, curr);
                    this._shipman.ships.push({ pos: [curr], length: len, posibleDirection: d });
                }
            }
        }
        return this;
    }
    checkValidity(curr, len, ship) {
        if (this.shipBefore(curr))
            return "no" /* no */;
        if (ship === undefined) {
            if (this.verticalCan(curr, len) && this.horizontalCan(curr, len))
                return "all" /* all */;
            else if (this.verticalCan(curr, len))
                return "vertical" /* vertical */;
            else if (this.horizontalCan(curr, len))
                return "horizontal" /* horizontal */;
            else
                return "no" /* no */;
        }
        else {
            for (let i of ship.pos) {
                if (Math.abs(curr - i) === this._gridSize && (ship.posibleDirection === "vertical" /* vertical */ || ship.posibleDirection === "all" /* all */))
                    return "vertical" /* vertical */;
                else if (Math.abs(curr - i) === 1 && (ship.posibleDirection === "horizontal" /* horizontal */ || ship.posibleDirection === "all" /* all */))
                    return "horizontal" /* horizontal */;
            }
            return "no" /* no */;
        }
    }
    //check if the the shippart can be placed vertically
    verticalCan(curr, limit) {
        limit--; //exclusive of the current so limit - 1;
        let top = 0;
        let t = curr;
        let b = curr;
        t = t - this._gridSize;
        while (t > 0) {
            if (this.shipBefore(t))
                break;
            top++;
            if (top >= limit)
                return true;
            t = t - this._gridSize;
        }
        let bottom = 0;
        b = b + this._gridSize;
        while (b <= this._gridSize * this._gridSize) {
            if (this.shipBefore(b))
                break;
            bottom++;
            if (bottom >= limit)
                return true;
            b = b + this._gridSize;
        }
        return top + bottom >= limit;
    }
    inHorizontalLine(curr, n) {
        if (curr % this._gridSize === 0)
            curr--;
        return Math.floor(curr / this._gridSize) * this._gridSize + 1 <= n && Math.floor(curr / this._gridSize) * this._gridSize + this._gridSize >= n;
    }
    horizontalCan(curr, limit) {
        limit--; //exclusive of the current so limit - 1;
        let right = 0;
        let r = curr;
        let l = curr;
        r = r + 1;
        while (this.inHorizontalLine(curr, r)) {
            if (this.shipBefore(r))
                break;
            right++;
            if (right >= limit)
                return true;
            r = r + 1;
        }
        let left = 0;
        l = l - 1;
        while (this.inHorizontalLine(curr, l)) {
            if (this.shipBefore(l))
                break;
            left++;
            if (left >= limit)
                return true;
            l = l - 1;
        }
        return left + right >= limit;
    }
    get shipman() {
        return this._shipman;
    }
}

class ComputerClass extends ShipClass {
    constructor(shipman, gridSize, shipsNo) {
        super(shipman, gridSize, shipsNo);
    }
    add(x) {
        let d;
        let len;
        if (this._shipman.ships.length !== 0) {
            if (this._shipman.ships[this._shipman.ships.length - 1].pos.length < this.shipsNo - this._shipman.ships.length + 1) {
                len = this.shipsNo - this._shipman.ships.length + 1;
                d = this.checkValidity(x, len, this._shipman.ships[this._shipman.ships.length - 1]);
            }
            else {
                len = this.shipsNo - this._shipman.ships.length;
                d = this.checkValidity(x, len);
            }
        }
        else {
            d = "all" /* all */;
            len = this.shipsNo;
        }
        this.addMultiple(x, d, len);
        return this;
    }
    addMultiple(x, d, len) {
        if (d === "no" /* no */)
            return;
        if (d === "all" /* all */)
            d = this.randomDirection();
        if (d === "horizontal" /* horizontal */)
            this.mulHori(x, len);
        else
            this.mulVert(x, len);
    }
    mulVert(x, len) {
        let xUp = x - this._gridSize;
        let xDown = x + this._gridSize;
        super.add(x);
        len--;
        let u = false;
        let d = false;
        while (len > 0) {
            if (!u)
                u = super.shipBefore(xUp);
            if (!d)
                d = super.shipBefore(xDown);
            if (xDown <= this._gridSize * this._gridSize && !d) {
                len--;
                super.add(xDown);
                xDown = xDown + this._gridSize;
            }
            if (len > 0 && xUp > 0 && !u) {
                len--;
                super.add(xUp);
                xUp = xUp - this._gridSize;
            }
        }
    }
    mulHori(x, len) {
        let xLeft = x - 1;
        let xRight = x + 1;
        super.add(x);
        let l = false;
        let r = false;
        len--;
        while (len > 0) {
            if (!l)
                l = super.shipBefore(xLeft);
            if (!r)
                r = super.shipBefore(xRight);
            if (super.inHorizontalLine(x, xLeft) && !l) {
                len--;
                super.add(xLeft);
                xLeft = xLeft - 1;
            }
            if (super.inHorizontalLine(x, xRight) && len > 0 && !r) {
                len--;
                super.add(xRight);
                xRight = xRight + 1;
            }
        }
    }
    randomDirection() {
        return ["vertical" /* vertical */, "horizontal" /* horizontal */][Math.floor(Math.random() * 2)];
    }
}

class ComputerShotClass extends ShotClass {
    constructor(oppShips) {
        super(oppShips);
        this._nextTry = {};
        for (let i = 0; i < this._oppShips.shipsNo; i++)
            this._nextTry[i + 1] = [];
        this._shipProcessing = [];
    }
    shoot(x, scoreBoard) {
        let s = 0; //aka shipLen
        //check for currently processing ship
        if (this._shipProcessing.length > 0) {
            for (let i = 0; i < this._shipProcessing.length; i++) {
                if (this._nextTry[this._shipProcessing[0]].length === this._shipProcessing[0]) { //prune out finished ships
                    this._shipProcessing.shift(); //no longer have to process that ship
                    i--;
                }
                else { //process that ship
                    s = this._shipProcessing[0];
                    //may require reaccessement of the vertical and horizontal viability
                    break;
                }
            } //after pruning out all the front ships that have aalrady been finished processing
            if (s > 0) { //that means it is processing from this._shipProcessing
                //just look at the last NextBest for that nextTry
                let nb = this._nextTry[s][this._nextTry[s].length - 1]; //last element of the nextTry[s]
                //check if it is still a single shot only
                if (this._nextTry[s].length < 2)
                    this.updateSingle(s, nb.hitVal, nb);
                for (let i = 0; i < nb.possible.length; i++) {
                    if (super.shotBefore(nb.possible[i])) {
                        nb.possible.shift();
                        i--; //so dont process that posible val
                    }
                    else { //never shot before and in the nextBest[best case scenerio]
                        super.shoot(nb.possible[i], scoreBoard);
                        if (this._shots[this._shots.length - 1].hit) {
                            if (this._shots[this._shots.length - 1].shipLen === s) {
                                this.updateNextTryMultiple(s, nb.possible[i]);
                            }
                            else { //missed the nextTry boat but coinsidently hit another boat
                                //OPTIMIZATION: check if it should go vertically or horizontally
                                this.updateNextTry(this._shots[this._shots.length - 1].shipLen, nb.possible.shift());
                            }
                        }
                        else {
                            nb.possible.shift();
                        }
                        break;
                    }
                }
            }
            else { //there is not more ships in queue to be processed so take randomShotAgain
                x = x = randomNum(this._oppShips.gridSize * this._oppShips.gridSize);
                while (!super.shoot(x, scoreBoard))
                    x = randomNum(this._oppShips.gridSize * this._oppShips.gridSize);
                if (this._shots[this._shots.length - 1].hit) { //this is a form of clean up
                    this.updateNextTry(this._shots[this._shots.length - 1].shipLen, x);
                }
            }
        }
        else {
            //random number
            x = x = randomNum(this._oppShips.gridSize * this._oppShips.gridSize);
            while (!super.shoot(x, scoreBoard))
                x = randomNum(this._oppShips.gridSize * this._oppShips.gridSize);
            if (this._shots[this._shots.length - 1].hit) { //this is a form of clean up
                //for the next try
                this.updateNextTry(this._shots[this._shots.length - 1].shipLen, x);
            }
        }
        return true;
    }
    addToShipProcessing(shipLen) {
        for (let i of this._shipProcessing) {
            if (i === shipLen)
                return; //dont push cuz already to be processed
        }
        this._shipProcessing.push(shipLen);
    }
    //x is the shot taken,
    /*These function can be in a helper file */
    horiGaps(nb) {
        let r = [];
        let tmp = nb.map((x) => x.hitVal).sort((a, b) => a < b ? -1 : a > b ? 1 : 0);
        //get all the nb.possible
        for (let i = 1; i < tmp.length; i++) {
            for (let j = 1; j < tmp[i] - tmp[i - 1]; j++) { //that means in between those two are the gap(s)
                r.push(tmp[i - 1] + j);
            }
        }
        return r;
    }
    horiPadding(nb, shipLen) {
        let r = [];
        let ht = nb.map((s) => s.hitVal).sort((a, b) => a < b ? -1 : a > b ? 1 : 0);
        ht = [ht[0], ht[ht.length - 1]];
        let bodyLen = ht[1] - ht[0];
        let right = ht[1] + 1;
        let left = ht[0] - 1;
        if (this.inHorizontalLine(ht[0], left) && !this.shotBefore(left) && !(ht[0] - left > shipLen - bodyLen))
            r.push(left);
        if (this.inHorizontalLine(ht[1], right) && !this.shotBefore(right) && !(right - ht[1] > shipLen - bodyLen))
            r.push(right);
        return r;
    }
    vertGaps(nb) {
        let ans = [];
        let tmp = nb.map((x) => x.hitVal).sort((a, b) => a < b ? -1 : a > b ? 1 : 0);
        for (let i = 1; i < tmp.length; i++) {
            for (let j = this._oppShips.gridSize; j < tmp[i] - tmp[i - 1]; j = j + this._oppShips.gridSize) { //that means in between those two are the gap(s)
                ans.push(tmp[i - 1] + j);
            }
        }
        return ans;
    }
    vertPadding(nb, shipLen) {
        //recommend one up(-gridSize) or one down(+gridSize) or both
        let ans = [];
        let ht = nb.map((s) => s.hitVal).sort((a, b) => a < b ? -1 : a > b ? 1 : 0);
        ht = [ht[0], ht[ht.length - 1]];
        let bodyLen = (ht[1] - ht[0]) / this._oppShips.gridSize;
        let down = ht[1] + this._oppShips.gridSize;
        let up = ht[0] - this._oppShips.gridSize;
        if (up > 0 && !super.shotBefore(up) && !(ht[0] - up > (shipLen - bodyLen) * this._oppShips.gridSize))
            ans.push(up);
        if (down <= this._oppShips.gridSize * this._oppShips.gridSize && !super.shotBefore(down) && !(down - ht[1] >= (shipLen - bodyLen) * this._oppShips.gridSize))
            ans.push(down);
        return ans;
    }
    updateNextTryMultiple(shipLen, x) {
        this._nextTry[shipLen].push({ hitVal: x, possible: [] });
        if (this._nextTry[shipLen].length === shipLen)
            return;
        let nb = this._nextTry[shipLen][this._nextTry[shipLen].length - 1];
        if (this.inHorizontalLine(this._nextTry[shipLen][0].hitVal, this._nextTry[shipLen][1].hitVal)) {
            let horiGaps = this.horiGaps(this._nextTry[shipLen]);
            if (horiGaps.length > 0) {
                nb.possible = horiGaps;
                return;
            }
            nb.possible = this.horiPadding(this._nextTry[shipLen], shipLen);
        }
        else {
            let vg = this.vertGaps(this._nextTry[shipLen]);
            if (vg.length > 0) {
                nb.possible = vg;
                return;
            }
            nb.possible = this.vertPadding(this._nextTry[shipLen], shipLen);
        }
    }
    updateNextTry(shipLen, x) {
        this.addToShipProcessing(shipLen);
        if (this._nextTry[shipLen].length > 0)
            this.updateNextTryMultiple(shipLen, x); //that means it got hit before perhaps accidently the last time
        else {
            let hori = this.calculateHori(shipLen, x);
            let vert = this.calculateVert(shipLen, x);
            if (hori.length > 0)
                this._nextTry[shipLen].push({ hitVal: x, possible: hori });
            else
                this._nextTry[shipLen].push({ hitVal: x, possible: vert });
        }
    }
    updateSingle(shipLen, x, nb) {
        //no need to put new next best just need to update the current nextbest
        let hori = this.calculateHori(shipLen, x);
        let vert = this.calculateVert(shipLen, x);
        if (hori.length > 0)
            nb.possible = hori;
        else
            nb.possible = vert;
    }
    calculateVert(shipLen, x) {
        // check if vertically is posible
        let upAns = [];
        let downAns = [];
        let upT = x - this._oppShips.gridSize;
        let downT = x + this._oppShips.gridSize;
        while (upT >= 1 && !this.shotBefore(upT) && ((x - upT) / this._oppShips.gridSize) + 1 <= shipLen) {
            upAns.push(upT);
            upT = upT - this._oppShips.gridSize;
        }
        while (downT <= this._oppShips.gridSize * this._oppShips.gridSize && !this.shotBefore(downT) && ((downT - x) / this._oppShips.gridSize) + 1 <= shipLen) {
            downAns.push(downT);
            downT = downT + this._oppShips.gridSize;
        }
        let ans = upAns.concat(downAns);
        if (ans.length + 1 < shipLen)
            return []; //not viable
        else {
            if (ans.length < 2)
                return ans;
            else {
                if (upAns.length > 0 && downAns.length > 0)
                    return [upAns[0], downAns[0]];
                else if (upAns.length > 0)
                    return [upAns[0]];
                else
                    return [downAns[0]];
            }
        }
    }
    calculateHori(shipLen, x) {
        //check how far it can go up
        // check if vertically is posible
        let leftAns = [];
        let rightAns = [];
        let leftT = x - 1;
        let rightT = x + 1;
        while (leftT >= 1 && this.inHorizontalLine(leftT, x) && !this.shotBefore(leftT) && (x - leftT + 1 <= shipLen)) {
            leftAns.push(leftT);
            leftT = leftT - 1;
        }
        while (rightT <= this._oppShips.gridSize * this._oppShips.gridSize && this.inHorizontalLine(rightT, x) && !this.shotBefore(rightT) && (rightT - x + 1 <= shipLen)) {
            rightAns.push(rightT);
            rightT = rightT + 1;
        }
        let ans = rightAns.concat(leftAns);
        if (ans.length + 1 < shipLen)
            return []; //not viable
        else {
            if (ans.length < 2)
                return ans;
            else {
                if (leftAns.length > 0 && rightAns.length > 0)
                    return [leftAns[0], rightAns[0]];
                else if (leftAns.length > 0)
                    return [leftAns[0]];
                else
                    return [rightAns[0]];
            }
        }
    }
    inHorizontalLine(curr, n) {
        if (curr % this._oppShips.gridSize === 0)
            curr--;
        return Math.floor(curr / this._oppShips.gridSize) * this._oppShips.gridSize + 1 <= n && Math.floor(curr / this._oppShips.gridSize) * this._oppShips.gridSize + this._oppShips.gridSize >= n;
    }
}

const initVal$ = (gridSize, shipsNo, iterations) => of({})
    .pipe(map(_ => {
    var cc = new ComputerClass({ ships: [], allPos: new Map() }, gridSize, shipsNo);
    return {
        "gridSize": gridSize,
        "shipsNo": shipsNo,
        board: cc,
        playerShot: new PlayerShotClass(cc),
        computerShot: new ComputerShotClass(cc)
    };
}));
const computerSetup$ = (sci, gridSize, ships) => interval().pipe(map(_ => Math.floor(Math.random() * (gridSize * gridSize)) + 1), scan((acc, curr) => acc.add(curr), sci), takeWhile((x) => x.shipman.ships.length < ships));
const ɵ0 = computerSetup$;
const setUp$ = (initial) => computerSetup$(initial.board, initial.gridSize, initial.shipsNo);

//initialize the pScore and cSccore()
//the shots$ will return [total computer shots, total player shots]
const shots$ = (b) => {
    function initScore() {
        let r = {};
        for (let i = 0; i < b.shipsNo; i++) {
            r[i + 1] = i + 1;
        }
        return r;
    }
    const pScore = new BehaviorSubject(initScore());
    const cScore = new BehaviorSubject(initScore());
    const oppBoard$ = of(b.board.shipman.ships);
    const computerShot$ = (c) => interval().pipe(takeWhile(_ => !(Object.values(cScore.value).filter((x) => x !== 0).length === 0)), scan((acc, curr) => ++acc, 0), tap(_ => c.shoot(0, cScore)), map((x) => {
        return { turns: x, shots: c.shots };
    }));
    const playerShot$ = (p) => interval().pipe(takeWhile(_ => !(Object.values(pScore.value).filter((x) => x !== 0).length === 0)), scan((acc) => ++acc, 0), tap(_ => p.shoot(0, pScore)), map((x) => {
        return { turns: x, shots: p.shots };
    }));
    return forkJoin({ oppBoard: oppBoard$, computerStatus: computerShot$(b.computerShot), playerStatus: playerShot$(b.playerShot) });
};

/*
map(_=>{
  if(Number.isNaN(shipsNo) || Number.isNaN(gridSize) || Number.isNaN(iterations))  throw new Battleship_performanceError("All parameters must not be NaN");
  if(shipsNo <=0 || gridSize <=0 || iterations <= 0)  throw new Battleship_performanceError("All parameters must be above 0");
  if(gridSize < shipsNo) throw new Battleship_performanceError("gridSize cannot be less than shipsNo");
  return;
}),
catchError((err:Battleship_performanceError)=> {
  return throwError(err)
})
*/
const checkNan = (n) => "" + n === "undefined";
const ɵ0$1 = checkNan;
const game$ = (gridSize, shipsNo, iterations) => of({}).pipe(tap(_ => {
    if (checkNan(shipsNo) || checkNan(gridSize) || checkNan(iterations))
        throw new Battleship_performanceError("All parameters must be filled");
    if (shipsNo <= 0 || gridSize <= 0 || iterations <= 0)
        throw new Battleship_performanceError("All parameters must be above 0");
    if (gridSize < shipsNo)
        throw new Battleship_performanceError("gridSize cannot be less than shipsNo");
}), catchError((err) => {
    return throwError(err);
}), concatMap(_ => initVal$(gridSize, shipsNo, iterations).pipe(switchMap((initial) => concat(setUp$(initial), shots$(initial))), last(), scan((acc, curr) => {
    acc.games++;
    acc.computerStatus = curr.computerStatus;
    acc.playerStatus = curr.playerStatus;
    acc.oppBoard = curr.oppBoard;
    return acc;
}, { games: 0, computerStatus: { turns: 0, shots: [] }, playerStatus: { turns: 0, shots: [] }, oppBoard: [] }), repeat(iterations))));

class BattleshipPerformanceService {
    game(gridSize, shipsNo, iterations) {
        return game$(gridSize, shipsNo, iterations);
    }
}
BattleshipPerformanceService.ɵprov = ɵɵdefineInjectable({ factory: function BattleshipPerformanceService_Factory() { return new BattleshipPerformanceService(); }, token: BattleshipPerformanceService, providedIn: "root" });
BattleshipPerformanceService.decorators = [
    { type: Injectable, args: [{
                providedIn: 'root'
            },] }
];

/*
 * Public API Surface of battleship-performance
 */

/**
 * Generated bundle index. Do not edit.
 */

export { BattleshipPerformanceService, Battleship_performanceError };
//# sourceMappingURL=battleship-performance.js.map
